#ifndef SHADE_H
#define SHADE_H
/****************************************************************************

FILE	shade.h

PURPOSE 
    	global declarations for structures, variables and functions.

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/


#include <stdio.h>


/* symbol flags */
#define S_PARAM         0x01
#define S_VAR           0x02
#define S_BUILTIN       0x04
#define S_REFERENCED    0x08
#define F_PARAM         0x10
#define F_VAR           0x20
#define F_RET           0x40
#define F_ILLUMIN       0x80
#define S_USED          0x100
#define S_OUTPUT        0x200
#define S_EXTERN        0x400

/* structure for symbols, defined in shader */
struct _sym {
    char      *name;
    int        type;
    char      *context;      /* context (func. or shader) name */
    int        offset;       /* offset in data vector          */
    int        block_count;  /* block nesting in current context for scope handling*/
    int        flags;
    int        dim;          /* variable dimension, used in arrays */
    int        repl_offset;  /* in func. parameters represents alias offset */
    int        reloc_offset;
    struct _sym *next;
};


#define MAXKIDS 4

/*
    List for backpatching 
*/
struct ilist {
    int i;                         /* command address */
    struct ilist *next_ilist;
};

struct patch {
    struct ilist *true, *false, *next_patch;
};


/* 
    Expression representation tree
*/
struct tree {
    int op;                      /* operation code     */
    int ty;                      /* result type        */
    int off;                     /* if leaf (represents symbol), then its offset */
    int ind;                     /* global node index, not used  */
    struct tree *next_tree;      /* not used           */
    struct tree *kids[MAXKIDS];  /* subexpressions     */
    struct patch tree_patch;
};


struct func_param{
        struct _sym *param;
        struct tree *array_index;       /* expression for index calculation */
        struct func_param * next_param;
};


/*
    Operation code structure
*/
struct ops{
    int op;                /* type          */
    char *name;            /* mnemonic name */
    int size;              /* operation size in bytes in machine code */
};

/*
    User defined function's structure
*/
struct func_tree {
    char * name;
    int    type;               /* return type */
    int    code_start;         /* preliminarily generated code start */
    int    code_end;           /* preliminarily generated code end   */
    int    off;                /* return value offset */
    int    rt_off;             /* offset during inlining for retval  alias */
    struct func_tree *next;    
    struct tree *func_code;    /* return expression */
    struct func_param *first_param; /* parameter list */
    int func_index;                 /* global index for function, not used */
    struct ilist * ret_patches;     /* return statement patches */
};

/*
    Global variables
*/
/* options flags */
extern int warning_flag;
extern int check_flag;
extern int output_functions;

/* input file parsing vars */
extern char *cur_input_file;
extern int lineno;

extern FILE * output_file;

extern int shaderstart;
extern char * shadername;
extern int shadertype;
extern unsigned char code[];



/* F U N C T I O N   P R O T O T Y P E S */

/* error handling functions */
void yyerror(char *s);
void undeclared(char *s);
void warning(const char *text);

/* code/data handling and output functions */
void s_fixcode(unsigned char *code, int size);
void s_fixvars(unsigned char *code, int size);
void s_dumpobj(unsigned char * code, int size);
void s_dumpcode(unsigned char * code, int size);
int calc_codesize(unsigned char * code, int size);

/* op code related functions */
struct ops* get_op(int op);
int is_branch_op(int op);
int is_offset_op(int op);
int is_string_op(int op);

/* expression tree functions */
struct tree *node2(int op, int ty, struct tree *p0, struct tree *p1);
struct tree *node3(int op, int ty,
        struct tree *p0, struct tree *p1, struct tree *p2);
struct tree *node4(int op, int ty,
        struct tree *p0, struct tree *p1, struct tree *p2, struct tree *p3);

struct tree *constnode(int ty, int offset);
int llen(struct tree *p); 
void notexpression(struct tree * expr);

/* outputting opcodes */
void emit(struct tree *t);
void emit1(unsigned char c);
void emit2(int op, int n);
void emit3(int op, int a, int b);
void emitshort(int n);
void storeshort(int n, int s);

/* symbol-handling related functions */
void sym_initialize(void);
void dumpsyms(void);
int install_fconst(double f);
int install_sconst(char *s);
int is_constant(int offset);
double getfloat(int offset);

char *getstring(int offset);
int get_string_offset();

struct _sym * install(char *s, int ty,  int off, int flags, int dim);
void add_active_symbol(struct _sym * sym);
void use_sym(struct _sym *s);
struct _sym *getsym(int offset);
struct _sym *lookup(char *s);
struct _sym *lookup_cur_context(char *s);
struct _sym *lookup_prev_decl(char *s);
struct _sym *get_string_sym(int offset);

int relocate_sym(int offset);
void end_sym_relocation(void);

int istriple(int type);
int isgeomtriple(int type);

/* functions, related to user function handling */
struct func_tree *get_function(char * name);
struct func_tree *get_function_by_offset(int offset);
struct func_tree * install_function(char * name, int type, int offset, int start, int end);
void add_param(struct func_tree *function, struct _sym *param);
void insert_function(char *name);
struct tree * fix_ret_expr(struct tree * function);
void add_ret_stmt(struct func_tree *function, struct ilist *stmt);


/* context-related functions */
char * get_context_name(void);
void get_context(int * function, int *block);
void set_context(char * context);
void remove_context(void);
void enter_block(void);
void leave_block(void);
int get_block(void);

/* illuminance related functions (creating/ending illuminance context) */
void start_illuminance(void);
void end_illuminance(void);
int in_illuminance(void);

/* support for continue and break statements */
void start_loop(void);
void end_loop(void);
void set_break(void);
void set_continue(void);
int reg_break(struct ilist * patch, int break_level);
int reg_continue(struct ilist *patch, int continue_level);

int get_dot(void);
void set_dot(int dot);
void backpatch(struct ilist * t, int n);


#endif
